# Copyright 2023 Emmanuel Chaboud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
from pathlib import Path


def absolute(path, root_dir=Path.cwd(), norm=False):
    """Helper function to convert a path to absolute.

    Unlike Path.resolve() and os.path.abspath, this function will not try to normalize by default the resulting path.

    Args:
        path: The path to transform.
        root_dir: The root directory, by default Path.cwd().
        norm: Should the path be normalized. False by default.
    Returns:
        The absolute path.
    """
    path_type = type(path)

    if not path.is_absolute():
        path = root_dir / path

    if norm:
        path = normalized(path)

    return path_type(path)


def normalized(path):
    """Helper function to normalize a path.

    This function is different from Path.resolve() because Path.resolve() will do actual filesystem access to resolve
    symlinks. This is more accurate but slower, and triggers https://bugs.python.org/issue35306 on Windows.

    This function instead use os.path.normpath to normalize the path, which does a simple string parsing that may be
    wrong if using Unix symlinks, but does not present the above issues.

    Args:
        path: The path to transform.
    Returns:
        The normalized path.
    """
    path_type = type(path)
    return path_type(os.path.normpath(path))


def find_common_prefix(files):
    """Find the common of prefix of the given file list.

    For example, common prefix of /tmp/file1.txt and /tmp/file2.txt is /tmp.

    Args:
        files: The file list.
    Returns:
        The path that it the common prefix of given files.
    """
    if len(files) == 0:
        return None

    all_parts = files[0].parts[:-1]
    common_parts = []

    for i, part in enumerate(all_parts):
        for f in files:
            if len(f.parts) - 1 <= i or f.parts[i] != part:
                return Path(*common_parts) if common_parts else None

        common_parts.append(part)

    return Path(*common_parts) if common_parts else None


def is_relative_to(path, *other):
    """Return True if the path is relative to another path or False.
    """
    try:
        path.relative_to(*other)
        return True
    except ValueError:
        return False


def with_stem(path, stem):
    """Return a new path with the stem changed.
    """
    return path.with_name(stem + path.suffix)


def append_stem(path, stem):
    """Return a new path with given stem added to the right of the current stem.
    """
    return with_stem(path, path.stem + stem)


def prepend_stem(path, stem):
    """Return a new path with given stem added to the left of the current stem.
    """
    return with_stem(path, stem + path.stem)


def append_name(path, name):
    """Return a new path with given name added to the right of the current name.
    """
    return path.with_name(path.name + name)


def prepend_name(path, name):
    """Return a new path with given name added to the left of the current name.
    """
    return path.with_name(name + path.name)



# Copyright 2023 Emmanuel Chaboud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from .unified_path import UnifiedPath


def make_path_validator(extensions=None):
    """Generate a path validator that optionaly checks the file extension.

    This function can be used for argparse argument checking:

    parser.add_argument(..., type=make_path_validator(extensions=['.json']), ...)

    Args:
        extensions: List of allowed extensions (including the .), by default None.
    Returns:
        The validator function.
    """

    def validator(obj):
        path = UnifiedPath(obj)

        if extensions is not None and path.suffix.lower() not in extensions:
            raise argparse.ArgumentTypeError('Path "%s" must have one of the following extensions: %s' %
                                             (obj, extensions))

        return path

    return validator



# Copyright 2023 Emmanuel Chaboud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import re

from .path_helper import is_relative_to

_DRIVE_TRANSLATIONS = []
_VIRTUAL_DRIVES = []

_DRIVE_REGEX = re.compile(r'^[a-zA-Z]:')
_UNC_REGEX = re.compile(r'^\\')
_POSIX_REGEX = re.compile(r'^/[^/]')


def register_drive_translation(nt_drive, posix_path):
    assert re.match(r'^[A-Z]:/', nt_drive), "nt_drive should be like C:/"
    assert _POSIX_REGEX.match(posix_path), "posix_path should be like /path"

    _DRIVE_TRANSLATIONS.append({'nt': nt_drive, 'posix': posix_path})


def register_virtual_drive(nt_drive, local_path):
    assert re.match(r'^[A-Z]:/', nt_drive), "nt_drive should be like C:/"

    _VIRTUAL_DRIVES.append({'nt': nt_drive, 'local': local_path})


class _CommonPosixPath(pathlib.PurePath):

    def as_unified(self):
        if self.is_absolute():
            # Check mappings
            for mapping in _VIRTUAL_DRIVES:
                if is_relative_to(self, mapping['local']):
                    return pathlib.PureWindowsPath(mapping['nt'], self.relative_to(mapping['local'])).as_posix()

            # Translate mount point
            mount = next((m for m in _DRIVE_TRANSLATIONS if is_relative_to(self, m['posix'])), None)
            if mount is not None:
                return pathlib.PureWindowsPath(mount['nt'], self.relative_to(mount['posix'])).as_posix()

        return os.fspath(self)


class _CommonWindowsPath(pathlib.PurePath):

    def as_unified(self):
        if self.is_absolute():
            # Check mappings
            for mapping in _VIRTUAL_DRIVES:
                if is_relative_to(self, mapping['local']):
                    return pathlib.PureWindowsPath(mapping['nt'], self.relative_to(mapping['local'])).as_posix()

        return self.as_posix()


class _PurePosixPath(pathlib.PurePosixPath, _CommonPosixPath):
    pass


class _PureWindowsPath(pathlib.PureWindowsPath, _CommonWindowsPath):
    pass


class _PosixPath(pathlib.PosixPath, _CommonPosixPath):
    pass


class _WindowsPath(pathlib.WindowsPath, _CommonWindowsPath):
    pass


class PureUnifiedPath(pathlib.PurePath):
    """A pathlib.PurePath that handles OS-independent path conversion.
    """

    def __new__(cls, *args):
        if len(args) == 0:
            return _PureWindowsPath() if os.name == 'nt' else _PurePosixPath()

        root = args[0]
        if isinstance(root, pathlib.PurePath):
            root = os.fspath(root)

        if not isinstance(root, str):
            raise TypeError('argument should be a str object or an os.PathLike '
                            'object returning str, not %r' % type(root))

        if _DRIVE_REGEX.match(root):
            # Drive letter and colon: Windows path
            path = _PureWindowsPath(*args)
        elif _UNC_REGEX.match(root):
            # UNC path
            path = _PureWindowsPath(*args)
        elif _POSIX_REGEX.match(root):
            # Slash followed by non-slash: Posix path
            path = _PurePosixPath(*args)
        else:
            # Relative path
            path = _PureWindowsPath(*args)

        # Check mappings
        if isinstance(path, pathlib.PureWindowsPath):
            if path.is_absolute():
                for mapping in _VIRTUAL_DRIVES:
                    if is_relative_to(path, mapping['nt']):
                        if os.name == 'nt':
                            return _PureWindowsPath(mapping['local'], path.relative_to(mapping['nt']))
                        else:
                            return _PurePosixPath(mapping['local'], path.relative_to(mapping['nt']))

        # Nt -> posix conversion
        if os.name == 'posix' and isinstance(path, pathlib.PureWindowsPath):
            if path.is_absolute():
                mount = next((m for m in _DRIVE_TRANSLATIONS if is_relative_to(path, m['nt'])), None)
                assert mount is not None, 'Cannot convert absolute NT path to posix: %s' % os.fspath(path)

                return _PurePosixPath(mount['posix'], path.relative_to(mount['nt']))

            return _PurePosixPath(path)

        # Posix -> nt conversion
        if os.name == 'nt' and isinstance(path, pathlib.PurePosixPath):
            if path.is_absolute():
                mount = next((m for m in _DRIVE_TRANSLATIONS if is_relative_to(path, m['posix'])), None)
                assert mount is not None, 'Cannot convert absolute posix path to NT: %s' % os.fspath(path)

                return _PureWindowsPath(mount['nt'], path.relative_to(mount['posix']))

            return _PureWindowsPath(path)

        # No conversion needed
        return path


class UnifiedPath(pathlib.Path):
    """A pathlib.Path that handles OS-independent path conversion.
    """

    def __new__(cls, *args):
        unified_path = PureUnifiedPath(*args)
        return _WindowsPath(unified_path) if os.name == 'nt' else _PosixPath(unified_path)



import os
from unittest.mock import patch

import pytest

import unified_path
from unified_path import PureUnifiedPath

root_dir = os.path.dirname(__file__)

unified_path.register_drive_translation('B:/', '/srv/database')
unified_path.register_virtual_drive('R:/', root_dir)


@patch('unified_path.os.name', 'posix')
def test_unified_path_posix():
    assert str(PureUnifiedPath('/srv/database/my/file')) == '/srv/database/my/file'
    assert str(PureUnifiedPath('B:/my/file')) == '/srv/database/my/file'
    assert str(PureUnifiedPath('B:\\my\\file')) == '/srv/database/my/file'

    assert str(PureUnifiedPath('/my/file')) == '/my/file'
    with pytest.raises(AssertionError):
        PureUnifiedPath('C:/my/file')
        PureUnifiedPath('C:\\my\\file')

    assert str(PureUnifiedPath('my/file')) == 'my/file'
    assert str(PureUnifiedPath('my\\file')) == 'my/file'


@patch('unified_path.os.name', 'nt')
def test_unified_path_nt():
    assert str(PureUnifiedPath('/srv/database/my/file')) == 'B:\\my\\file'
    assert str(PureUnifiedPath('B:/my/file')) == 'B:\\my\\file'
    assert str(PureUnifiedPath('B:\\my\\file')) == 'B:\\my\\file'

    assert str(PureUnifiedPath('C:/my/file')) == 'C:\\my\\file'
    assert str(PureUnifiedPath('C:\\my\\file')) == 'C:\\my\\file'
    with pytest.raises(AssertionError):
        PureUnifiedPath('/my/file')

    assert str(PureUnifiedPath('my/file')) == 'my\\file'
    assert str(PureUnifiedPath('my\\file')) == 'my\\file'


def test_unified_path_unified():
    assert PureUnifiedPath('/srv/database/my/file').as_unified() == 'B:/my/file'
    assert PureUnifiedPath('B:/my/file').as_unified() == 'B:/my/file'
    assert PureUnifiedPath('B:\\my\\file').as_unified() == 'B:/my/file'

    assert PureUnifiedPath('my/file').as_unified() == 'my/file'
    assert PureUnifiedPath('my\\file').as_unified() == 'my/file'


def test_unified_path_virtual_drive():
    assert str(PureUnifiedPath('R:/my/file')) == str(PureUnifiedPath(root_dir, 'my/file'))
    assert PureUnifiedPath(root_dir, 'my/file').as_unified() == 'R:/my/file'



# Copyright 2023 Emmanuel Chaboud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import PurePath

from .unified_path import UnifiedPath


class UnifiedPathEncoder(json.JSONEncoder):
    """A JSON encoder that serializes a Path to the string corresponding to the unified representation,
    that is OS-independent.
    """

    def default(self, obj):
        if isinstance(obj, PurePath):
            return UnifiedPath(obj).as_unified()

        return json.JSONEncoder.default(self, obj)



from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(name='unified-path',
      version='1.0',
      description='A Python wrapper around pathlib providing an OS-independent path representation.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Emmanuel Chaboud',
      url='https://github.com/emmcb/unified-path',
      classifiers=[
          'License :: OSI Approved :: Apache Software License', 'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3 :: Only', 'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      python_requires='>=3')
